@@ -1,7 +1,8 @@
 import numpy as np
-
+import triangle as tr
 from scipy.spatial import Delaunay as delaunay
-from shapely.geometry.polygon import Polygon as ShapelyPolygon
+from shapely.geometry import Polygon as ShapelyPolygon
+from shapely.geometry import LineString as ShapelyLine
 from shapely.geometry import Point as ShapelyPoint
 
 class Mesh:
@@ -106,47 +107,60 @@ def prepareGeometryForMeshing(shapes):
         else:
             shapes[i].isChild = True
 
-def generateNodes(shapes, crse, meshingMethod):
+def generateNodes(shapes, meshingMethod, maxAngle, maxArea, crse):
     # Generate interior free nodes
-    for i in range(len(shapes)):
-        if shapes[i].isClosed:
-            if meshingMethod == "ruppert":
+    if meshingMethod == "ruppert":
+        for i in range(len(shapes)):
+            if shapes[i].isClosed:
                 meshShapeRuppert(i, shapes, crse)
-            elif meshingMethod == "triangle":
-                meshShapeTriangle(i, shapes, crse)
-        else:
-            shapes[i].freeNodes = np.array([])
+            else:
+                shapes[i].freeNodes = np.array([])
+    elif meshingMethod == "triangle":
+        meshShapesTriangle(shapes,maxAngle, maxArea)
 
-def meshShapeTriangle(ind, shapes):
-    # Parent points
-    parentPoints = shapes[ind].boundaryNodes
-    # Child points
-    childPoints = []
-    for i in range(len(shapes[ind].child)):
-        childPoints.append(shapes[shapes[ind].child[i]].boundaryNodes.tolist())
-    
-    vertices = parentPoints.append(childPoints)
-
-    # Parent segments
-    segments = []
-    for i in range(len(parentPoints)-1):
-        segments.append([i, i+1])
-    segments.append([len(parentPoints)-1, 0])
-    # Child segments
-    segInd = len(parentPoints)
-    for i in range(len(shapes[ind].child)):
-        for j in range(len(childPoints[i])-1):
-            if shapes[shapes[ind].child[i]].shape != "point":
-                segments.append([segInd+j, segInd+j+1])
-        if shapes[shapes[ind].child[i]].isClosed:
-            segments.append([segInd+j+1, segInd])
-        if shapes[shapes[ind].child[i]].shape == 'point':
-            segInd = segInd + 1
-        else:
-            segInd = segInd + len(childPoints[i])
-
-    regions = []
+def meshShapesTriangle(shapes, maxAngle, maxArea):
+    iv, fv = triangleMesh(shapes, maxAngle, maxArea)
+    initNodes = iv.shape[0]
+    finalNodes = fv.shape[0]
+    for shape in shapes:
+        shape.freeNodes = np.empty((0,2))
+    for i in range(initNodes,finalNodes):
+        extraVertex = ShapelyPoint(fv[i][0],fv[i][1])
+        for shape in shapes:
+            if shape.isClosed:
+                shapePolygon = ShapelyPolygon(shape.boundaryNodes.tolist())
+                # if extraVertex is on boundary
+                if not shapePolygon.contains(extraVertex) and shapePolygon.covers(extraVertex):
+                    shape.boundaryNodes = np.vstack([shape.boundaryNodes, np.array([fv[i][0],fv[i][1]])])
+                # if extraVertex is inside
+                if shapePolygon.contains(extraVertex):
+                    vertexOutside = True
+                    for j in shape.child:
+                        childShape = shapes[j]
+                        childPolygon = ShapelyPolygon(childShape.boundaryNodes.tolist())
+                        if childPolygon.covers(extraVertex):
+                            vertexOutside = False
+                            break
+                    if vertexOutside:
+                        shape.freeNodes = np.vstack([shape.freeNodes, np.array([fv[i][0],fv[i][1]])])
+            elif shape.shape != 'point':
+                lineString = ShapelyLine(shape.boundaryNodes.tolist())
+                if lineString.contains(extraVertex):
+                    shape.boundaryNodes = np.vstack([shape.boundaryNodes, np.array([fv[i][0],fv[i][1]])])
 
+def triangleMesh(shapes, maxAngle = "", maxArea = ""):
+    nov = 0
+    ver = np.empty((0,2))
+    seg = np.empty((0,2), dtype=int)
+    for shape in shapes:
+        ver = np.vstack([ver, shape.boundaryNodes])
+        if shape.shape != 'point':
+            seg = np.vstack([seg, np.asarray(shape.segments, dtype=int)+ nov])
+        nov = ver.shape[0]
+    pslg = dict(vertices=ver, segments=seg)
+    stri = "pq"+maxAngle+"a"+maxArea+"A"
+    tri = tr.triangulate(pslg, stri)
+    return ver, tri['vertices']
 
 def meshShapeRuppert(ind, shapes, crse): # Consider skipping sengment initialisation in case of Point
     # Parent points
